#ifndef __ARCLIB_SAS_H__
#define __ARCLIB_SAS_H__

#include "commdef.h"

#include "if.h"

#include "iostream"
#include "vector"

#if defined(_WIN32)
#include <WinSock2.h>
#endif

#include "sasbase.h"
#include "fibrebase.h"
#include "scsibase.h"

#ifdef __cplusplus
extern "C" {
#endif


#define ARC_BIT(x)	(1L << x)
#define MV_BIT(x) ARC_BIT(x)
typedef unsigned int	U32;
typedef unsigned char	U8;
typedef unsigned short	U16;
typedef unsigned char   BYTE;


#if defined(_WIN32)
	typedef unsigned __int64 _U64;
	typedef   signed __int64 _I64;
#else
	typedef unsigned long long _U64;
	typedef   signed long long _I64;
#endif



typedef union 
{
	struct 
	{
		U32 low;
        U32 high;
    } parts;
    _U64 value;
} U64, *PU64;



#if defined(__linux__) || defined(__FreeBSD__) || defined(__sun__) || defined(__APPLE__)
// Linux/FreeBSD has strcasecmp function but do not have stricmp(Windows)
// so we must map the stricmp to the strcasecmp in the linux/FreeBSD system
	#define	stricmp strcasecmp
	
	typedef bool BOOLEAN;
	#define UNIX_LIKE_OS	1

#elif defined(_WIN32)

#elif defined(_DOS)
#endif


// flags of getting event
#define FLAG_ALL			0
#define FLAG_NEW			1

typedef union sTIMESTAMP
{
	/* 32 bits */
	struct
	{
		DWORD tmSecond:6; /* LSB - second */
		DWORD tmMinute:6;
		DWORD tmMonth:4;
		DWORD tmHour:5; // --> 5
		DWORD tmDate:5;
		DWORD tmYear:5;	 /* MSB - year */
		DWORD tmFlag:1;
	} u;
	struct
	{
		DWORD tmTick:31;
		DWORD tmFlag:1;
	} x;
} sTIMESTAMP, *pTIMESTAMP;

/****************************************************************************************/
#define ARCSAS_MAX_WIDEPORT_PHYS               4
#define ARCSAS_MAX_EXPANDER_SUPPORTED          8
#define ARCSAS_MAX_ENCLOSURE_SUPPORTED         8
#define ARCSAS_MIN_DEVICE_ID				   0
#define ARCSAS_MAX_DEVICE_ID                   135
#define ARCSAS_MIN_HDD_ID				       ARCSAS_MIN_DEVICE_ID
#define ARCSAS_MAX_HDD_ID                      127
#define ARCSAS_MIN_ENCLOSURE_ID                128 
#define ARCSAS_MAX_ENCLOSURE_ID		           ARCSAS_MAX_DEVICE_ID  
#define ARCSAS_MIN_EXPANDER_ID                 136  
#define ARCSAS_MAX_EXPANDER_ID			       143  
#define ARCSAS_MIN_PM_ID				       144 
#define ARCSAS_MAX_PM_ID				       151
#define ARCSAS_MAX_HDD_ELEMENT                 (ARCSAS_MAX_HDD_ID-ARCSAS_MIN_HDD_ID+1)
#define ARCSAS_MAX_ENCLOSURE_ELEMENT           (ARCSAS_MAX_ENCLOSURE_SUPPORTED+1) /* add one internal enclosure */
#define ARCSAS_MAX_EXPANDER_ELEMENT            (ARCSAS_MAX_EXPANDER_SUPPORTED)
#define ARCSAS_MAX_PM_ELEMENT                  (ARCSAS_MAX_PM_ID-ARCSAS_MIN_PM_ID+1)
#define ARCSAS_MAX_EVENT_ELEMENT               128  /* event records */
#define ARCSAS_ID_NOT_MAPPED			       0xff

#define ARCSAS_MIN_TARGETLUN                   0
#define ARCSAS_MAX_TARGETLUN                   127

#define ARCSAS_VIRTUAL_SES_DEVICE              136

#define MAX_BUFFER_LEN                         4096 + 64

/*******************/
/* CDB definitions */
/*******************/


#define CDB_LEN				16
#define API_CDB0_ADAPTER_COMMAND                                0xF0
    /* 
    *********for Adapter CDB[1] ***************
    */
    #define API_CDB1_ADAPTER_GETCOUNT                                   0
    #define API_CDB1_ADAPTER_GETINFO                                    (API_CDB1_ADAPTER_GETCOUNT + 1)
    #define API_CDB1_ADAPTER_GETCONFIG                                  (API_CDB1_ADAPTER_GETCOUNT + 2)
    #define API_CDB1_ADAPTER_SETCONFIG                                  (API_CDB1_ADAPTER_GETCOUNT + 3)
    #define API_CDB1_ADAPTER_POWER_STATE_CHANGE                         (API_CDB1_ADAPTER_GETCOUNT + 4)
    #define API_CDB1_ADAPTER_SET_PASSWORD                               (API_CDB1_ADAPTER_GETCOUNT + 5)
    #define API_CDB1_ADAPTER_DO_DEVICES_DISCOVER                        (API_CDB1_ADAPTER_GETCOUNT + 6)
    #define API_CDB1_ADAPTER_MAX                                        (API_CDB1_ADAPTER_GETCOUNT + 7)

#define API_CDB0_LOGICAL_DISK_COMMAND                           0xF1
    /* 
    *********for logical disk CDB[1]*********** 
    */
    #define API_CDB1_LOGICAL_DISK_CREATE                                0
    #define API_CDB1_LOGICAL_DISK_GETMAXSIZE                            (API_CDB1_LOGICAL_DISK_CREATE + 1)
    #define API_CDB1_LOGICAL_DISK_GETINFO                               (API_CDB1_LOGICAL_DISK_CREATE + 2)
    #define API_CDB1_LOGICAL_DISK_GETTARGETLDINFO                       (API_CDB1_LOGICAL_DISK_CREATE + 3)
    #define API_CDB1_LOGICAL_DISK_DELETE                                (API_CDB1_LOGICAL_DISK_CREATE + 4)
    #define API_CDB1_LOGICAL_DISK_GETSTATUS                             (API_CDB1_LOGICAL_DISK_CREATE + 5)
    #define API_CDB1_LOGICAL_DISK_GETCONFIG                             (API_CDB1_LOGICAL_DISK_CREATE + 6)
    #define API_CDB1_LOGICAL_DISK_SETCONFIG                             (API_CDB1_LOGICAL_DISK_CREATE + 7)
    #define API_CDB1_LOGICAL_DISK_STARTREBUILD                          (API_CDB1_LOGICAL_DISK_CREATE + 8)
    #define API_CDB1_LOGICAL_DISK_STARTCONSISTENCYCHECK                 (API_CDB1_LOGICAL_DISK_CREATE + 9)
    #define API_CDB1_LOGICAL_DISK_STARTINIT                             (API_CDB1_LOGICAL_DISK_CREATE + 10)
    #define API_CDB1_LOGICAL_DISK_STARTMIGRATION                        (API_CDB1_LOGICAL_DISK_CREATE + 11)
    #define API_CDB1_LOGICAL_DISK_BGACONTROL                            (API_CDB1_LOGICAL_DISK_CREATE + 12)
    #define API_CDB1_LOGICAL_DISK_WIPEMDD                               (API_CDB1_LOGICAL_DISK_CREATE + 13)
    #define API_CDB1_LOGICAL_DISK_GETSPARESTATUS                        (API_CDB1_LOGICAL_DISK_CREATE + 14)
    #define API_CDB1_LOGICAL_DISK_SETGLOBALSPARE                        (API_CDB1_LOGICAL_DISK_CREATE + 15)
    #define API_CDB1_LOGICAL_DISK_SETLDSPARE                            (API_CDB1_LOGICAL_DISK_CREATE + 16)
    #define API_CDB1_LOGICAL_DISK_REMOVESPARE                           (API_CDB1_LOGICAL_DISK_CREATE + 17)
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_SETSTATUS               (API_CDB1_LOGICAL_DISK_CREATE + 18)
    #define API_CDB1_LOGICAL_DISK_SHUTDOWN                              (API_CDB1_LOGICAL_DISK_CREATE + 19)
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_FREE_SPACE_INFO         (API_CDB1_LOGICAL_DISK_CREATE + 20)
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_GETMBRINFO              (API_CDB1_LOGICAL_DISK_CREATE + 21)
    #define API_CDB1_LOGICAL_DISK_SIZEOF_MIGRATE_TARGET                 (API_CDB1_LOGICAL_DISK_CREATE + 22)
    #define API_CDB1_LOGICAL_DISK_TARGET_LUN_TYPE			            (API_CDB1_LOGICAL_DISK_CREATE + 23)
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_MPCHECK                 (API_CDB1_LOGICAL_DISK_CREATE + 24)
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_GETMPSTATUS             (API_CDB1_LOGICAL_DISK_CREATE + 25)
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_GET_RCT_COUNT           (API_CDB1_LOGICAL_DISK_CREATE + 26)
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_RCT_REPORT              (API_CDB1_LOGICAL_DISK_CREATE + 27)
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_START_DATASCRUB         (API_CDB1_LOGICAL_DISK_CREATE + 28)	// temp
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_BGACONTROL	            (API_CDB1_LOGICAL_DISK_CREATE + 29)	// temp
    #define API_CDB1_LOGICAL_DISK_PHYSICAL_DISK_GETBGASTATUS            (API_CDB1_LOGICAL_DISK_CREATE + 30)	// temp
    // Added reserved for future expansion so that MRU/CLI can be backward compatible with drier.
    #define API_CDB1_LOGICAL_DISK_RESERVED1				                (API_CDB1_LOGICAL_DISK_CREATE + 31)	
    #define API_CDB1_LOGICAL_DISK_RESERVED2				                (API_CDB1_LOGICAL_DISK_CREATE + 32)	
    #define API_CDB1_LOGICAL_DISK_RESERVED3				                (API_CDB1_LOGICAL_DISK_CREATE + 33)	
    #define API_CDB1_LOGICAL_DISK_RESERVED4				                (API_CDB1_LOGICAL_DISK_CREATE + 34)	
    #define API_CDB1_LOGICAL_DISK_RESERVED5				                (API_CDB1_LOGICAL_DISK_CREATE + 35)	
    #define API_CDB1_LOGICAL_DISK_MAX                                   (API_CDB1_LOGICAL_DISK_CREATE + 36)

#define API_CDB0_PHYSICAL_DISK_COMMAND                          0xF2
    /* 
    **********for Block device CDB[1]***********
    */
    #define API_CDB1_PHYSICAL_DISK_GET_INFO                             0
    #define API_CDB1_PHYSICAL_DISK_BLOCKIDS                             1
    #define API_CDB1_PHYSICAL_DISK_MAX                                  2

#define API_CDB0_PERIPHERAL_DEVICE_COMMAND                      0xF3
    /* 
    *********for peripheral device CDB[1], CDB[2]: "device entry index".. SD_ID or EXP_ID PM_ID
    */
    #define API_CDB1_PERIPHERAL_DEVICE_GET_SCSI_DEVICE_INFO             0   /*CDB[2]:scsi id(0...135),CDB[3]:lun id*/
    #define API_CDB1_PERIPHERAL_DEVICE_GET_EXPANDER_INFO                1   /*CDB[2]:0...7  */
    #define API_CDB1_PERIPHERAL_DEVICE_GET_PM_INFO                      2   /*CDB[2]:0...7  */
    #define API_CDB1_PERIPHERAL_DEVICE_GET_DISK_CONFIG                  3
    #define API_CDB1_PERIPHERAL_DEVICE_SET_DISK_CONFIG                  4
        /* 
        *********Sub command for API_CDB1_PERIPHERAL_DEVICE_SET_DISK_CONFIG 
        */
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_WRITE_CACHE_OFF     0
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_WRITE_CACHE_ON      1
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_SMART_OFF           2
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_SMART_ON            3
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_DATA_PROTECT        4
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_SPEED_1_5G	        5
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_SPEED_3G	        6
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_SPEED_6G	        7
        #define API_CDB4_PERIPHERAL_DEVICE_SET_DISK_ID                  8

    #define API_CDB1_PERIPHERAL_DEVICE_BSL_DUMP                         5
    #define API_CDB1_PERIPHERAL_DEVICE_GET_LUN_DEVICE_INFO  	        6	// not used
    #define API_CDB1_PERIPHERAL_DEVICE_RESERVED1				        7	// not used
    #define API_CDB1_PERIPHERAL_DEVICE_GET_SMART_STATUS                 8
        /* 
        *********Sub command for API_CDB1_PERIPHERAL_DEVICE_GET_SMART_STATUS 
        */
        #define API_CDB4_PERIPHERAL_DEVICE_SMART_RETURN                 0
    #define API_CDB1_PERIPHERAL_DEVICE_MAX                              9

#define API_CDB0_EVENT_COMMAND                                  0xF4
    /* 
    *********************for event********************** 
    */
    #define API_CDB1_EVENT_GET_ALL_EVENT                                0
    #define API_CDB1_EVENT_GET_CURRENT_EVENT_INDEX                      1
	#define API_CDB1_EVENT_CLEAR_ALL_EVENT                              2

#define API_CDB0_DEBUG_COMMAND                                  0xF5
    /* 
    *********************for DBG ***********************
    */
    #define API_CDB1_DBG_PDWR                                           0
    #define API_CDB1_DBG_MAP                                            1
    #define API_CDB1_DBG_MAX                                            2

#define API_CDB0_FLASH_COMMAND                                  0xF6
    /* 
    *********************for FLASH *********************
    */
    #define API_CDB1_GET_BIOS_FLASH_SIZE                                0
    #define API_CDB1_READ_BIOS_FLASH_OFFSET                             1
    #define API_CDB1_ERASE_BIOS_FLASH                                   2
    #define API_CDB1_WRITE_BIOS_FLASH_OFFSET                            3
    #define API_CDB1_SYNC_FLASH_DEVICES_MAP_WITH_SLOT_NUM               4
    #define API_CDB1_FLASH_MAX                                          5

/* for SDI(HP CSMI) */
#define API_CDB0_CSMI_CORE_COMMAND                              0xF7
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_GENERAL                                       0x00
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_MANUFACTURER_INFORMATION                      0x01
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_READ_GPIO_REGISTER                                   0x02
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_SELF_CONFGRIGRUATION_STATUS                   0x03
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_DISCOVER                                             0x10
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_PHY_ERROR_LOG                                 0x11
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_PHY_SATA                                      0x12
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_ROUTE_INFORMATION                             0x13
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_PHY_EVENT_INFORMATION                         0x14
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_PHY_BROADCAST_COUNTS                          0x15
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_DISCOVER_LIST                                        0x16
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_REPORT_EXPANDER_ROUTE_TABLE                          0x17
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_CONFIGURE_GENERAL                                    0x80
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_ENABLE_DISABLE_ZONING                                0x81
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_WRITE_GPIO_REGISTER                                  0x82
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_ZONED_BROADCAST                                      0x85
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_CONFIGURE_ROUTE_INFORMATION                          0x90
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL                                          0x91
    #define API_CDB5_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL_PHY_NOP                                  0x00
    #define API_CDB5_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL_LINK_RESET                               0x01
    #define API_CDB5_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL_HARD_RESET                               0x02
    #define API_CDB5_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL_DISABLE                                  0x03
    #define API_CDB5_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL_CLEAR_ERROR_LOG                          0x05
    #define API_CDB5_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL_CLEAR_AFFILIATION                        0x06
    #define API_CDB5_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL_TRANSMIT_SATA_PORT_SELECTION_SIGNAL      0x07
    #define API_CDB5_CSMI_SAS_SMP_PASSTHRU_PHY_CONTROL_CLEAR_STP_NEXUS_LOSS                     0x08
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_PHY_TEST                                             0x92
    #define API_CDB4_CSMI_SAS_SMP_PASSTHRU_CONFIGURE_PHY_EVENT_INFORMATION                      0x93

#define API_CDB0_CSMI_RAID_COMMAND                              0xF8

#define API_CDB0_PASS_THRU_SCSI_COMMAND	                        0xFA
    #define API_CDB1_SCSI_NON_DATA					                0x00
    #define API_CDB1_SCSI_DATA_IN						            0x01
    #define API_CDB1_SCSI_DATA_OUT					                0x02

#define API_CDB0_PASS_THRU_ATA_COMMAND			                0xFB
    #define API_CDB1_ATA_NON_DATA					                0x00
    #define API_CDB1_ATA_DATA_IN						            0x01
    #define API_CDB1_ATA_DATA_OUT			                        0x02

#define API_CDB0_AES                                            0xFC
    #define API_CDB1_AES_SET_LINK                                   0x00
    #define API_CDB1_AES_GETCONFIG                                  0x01
    #define API_CDB1_AES_SETCONFIG                                  0x02



#define GLOBAL_SPARE_DISK                       ARC_BIT(2)

#define PD_DDF_VALID                            ARC_BIT(0)
#define PD_DISK_VALID                           ARC_BIT(1)
#define PD_DDF_CLEAN                            ARC_BIT(2)
#define PD_NEED_UPDATE                          ARC_BIT(3)
#define PD_MBR_VALID                            ARC_BIT(4)
#define PD_NEED_FLUSH                           ARC_BIT(5)
#define PD_CLEAR_MBR                            ARC_BIT(6)
#if ERROR_HANDLING_SUPPORT
#define PD_RCT_NEED_UPDATE             			ARC_BIT(7)
#endif

#define PD_STATE_ONLINE                         ARC_BIT(0)
#define PD_STATE_FAILED                         ARC_BIT(1)
#define PD_STATE_REBUILDING                     ARC_BIT(2)
#define PD_STATE_TRANSITION                     ARC_BIT(3)
#define PD_STATE_SMART_ERROR                    ARC_BIT(4)
#define PD_STATE_READ_ERROR                     ARC_BIT(5)
#define PD_STATE_MISSING                        ARC_BIT(6)

#define EXP_SSP                                 ARC_BIT(0)
#define EXP_STP                                 ARC_BIT(1)
#define EXP_SMP                                 ARC_BIT(2)

#define SUPPORT_LD_MODE_RAID0                   ARC_BIT(0)
#define SUPPORT_LD_MODE_RAID1                   ARC_BIT(1)
#define SUPPORT_LD_MODE_RAID10                  ARC_BIT(2)
#define SUPPORT_LD_MODE_RAID1E                  ARC_BIT(3)
#define SUPPORT_LD_MODE_RAID5                   ARC_BIT(4)
#define SUPPORT_LD_MODE_RAID6                   ARC_BIT(5)
#define SUPPORT_LD_MODE_RAID50                  ARC_BIT(6)
#define SUPPORT_LD_MODE_JBOD                    ARC_BIT(7)

#define FEATURE_BGA_REBUILD_SUPPORT             ARC_BIT(0)
#define FEATURE_BGA_BKINIT_SUPPORT				ARC_BIT(1)
#define FEATURE_BGA_SYNC_SUPPORT				ARC_BIT(2)
#define FEATURE_BGA_MIGRATION_SUPPORT           ARC_BIT(3)
#define FEATURE_BGA_MEDIAPATROL_SUPPORT         ARC_BIT(4)


#pragma pack(1)


// DevType
#define DEVICE_TYPE_NONE                        0
#define DEVICE_TYPE_HD                          1		// DT_DIRECT_ACCESS_BLOCK
#define DEVICE_TYPE_PM                          2
#define DEVICE_TYPE_EXPANDER                    3		// DT_EXPANDER
#define DEVICE_TYPE_TAPE						4		// DT_SEQ_ACCESS
#define DEVICE_TYPE_PRINTER						5		// DT_PRINTER
#define DEVICE_TYPE_PROCESSOR					6		// DT_PROCESSOR
#define DEVICE_TYPE_WRITE_ONCE					7 		// DT_WRITE_ONCE
#define DEVICE_TYPE_CD_DVD						8		// DT_CD_DVD
#define DEVICE_TYPE_OPTICAL_MEMORY				9 		// DT_OPTICAL_MEMORY
#define DEVICE_TYPE_MEDIA_CHANGER				10		// DT_MEDIA_CHANGER
#define DEVICE_TYPE_ENCLOSURE					11		// DT_ENCLOSURE
#define DEVICE_TYPE_PORT                        0xFF	// DT_STORAGE_ARRAY_CTRL
typedef struct _Link_Endpoint 
{
	U16      DevID;
	U8       DevType;						/* Refer to DEVICE_TYPE_xxx, (additional type like EDGE_EXPANDER and 
											   FANOUT_EXPANDER might be added). */
	U8       PhyCnt;						/* Number of PHYs for this endpoint.  Greater than 1 if it is wide port. */
	U8       PhyID[ARCSAS_MAX_WIDEPORT_PHYS];		/* Assuming wide port has max of 'MAX_WIDEPORT_PHYS' PHYs. */
	U8       SAS_Address[8];				/* Filled with 0 if not SAS device. */
	U8       Reserved[12];
} Link_Endpoint, * PLink_Endpoint;			/* total 28 bytes */

typedef struct _Link_Entity 
{
	Link_Endpoint    Parent;
	U8				 Reserved[8];
	Link_Endpoint    Self;
} Link_Entity,  *PLink_Entity;				/* total 64 bytes */

typedef struct _Version_Info
{
	U32        VerMajor;
	U32        VerMinor;
	U32        VerOEM;
	U32        VerBuild;
}Version_Info, *PVersion_Info;

typedef struct _Adapter_Info 
{
	Version_Info    DriverVersion;
	Version_Info    BIOSVersion;
	U64          Reserved1[2];				/* Reserve for firmware */

	U32          SystemIOBusNumber;
	U32          SlotNumber;
	U32          InterruptLevel;
	U32          InterruptVector;
    
	U16          VenID;
	U16          SubVenID;
	U16          DevID;
	U16          SubDevID;
    
	U8           PortCount;					/* How many ports, like 4 ports, or 4S1P. */
	U8           PortSupportType;			/* Like SATA port, SAS port, PATA port, use ARC_BIT */
	U8           Features;					/* Feature bits.  See FEATURE_XXX. If corresponding bit is set, that feature is supported. */
	BOOLEAN      AlarmSupport;
	U8           RevisionID;				/* Chip revision */
	U8           Reserved2[11];
    
	U8           MaxTotalBlocks;    
	U8           MaxBlockPerPD;
	U8           MaxHD;
	U8           MaxExpander;
	U8           MaxPM;
	U8           MaxLogicalDrive;
	U16          LogicalDriverMode;

	U8           WWN[8];					/* For future VDS use. */
} Adapter_Info, *PAdapter_Info;				/* total 104 bytes */


#define HBA_FLAG_INT13_ENABLE               ARC_BIT(0)   // int 13h enable/disable
#define HBA_FLAG_SILENT_MODE_ENABLE         ARC_BIT(1)
#define HBA_FLAG_FIXED_MAP_ENABLE           ARC_BIT(2)   // fixed physical/logical mapping enable/disable
#define HBA_FLAG_TO_CLEAR_MAP               ARC_BIT(3)   // persistent device map is to be cleared
#define HBA_FLAG_ALL_DISKS_WRITE_PROTECT    ARC_BIT(4)   // disk write protection enabled/disable
#define HBA_FLAG_SATA_NCQ_ENABLE            ARC_BIT(5)   // SATA drive NCQ enable/disable
#define HBA_FLAG_SMART_STATUS_POLL_ENABLE   ARC_BIT(6)   // SMART status polling enable/disable
#define HBA_FLAG_DISK_WRITE_CACHE_ENABLE    ARC_BIT(7)   // disk write cache enable/disable
#define HBA_FLAG_MAX_PHY_LINK_RATE          (ARC_BIT(10) | ARC_BIT(9) | ARC_BIT(8))   // maximum phy link rate - bit8~10(3 bits)
#define HBA_FLAG_MCXL_4M                    ARC_BIT(11)  // maximum command xfer length = 4MB - bit11

#define DEFAULT_PARAMETERS                  (HBA_FLAG_INT13_ENABLE |                    \
                                            HBA_FLAG_SATA_NCQ_ENABLE |                  \
                                            HBA_FLAG_DISK_WRITE_CACHE_ENABLE |          \
                                            HBA_FLAG_MPL_AUTO |                         \
                                            HBA_FLAG_MCXL_4M)



typedef struct _Adapter_Config 
{
    U32     BIOS_Flag;/*4 bytes*/
    U32     Reserved; /*4 bytes*/
} Adapter_Config, *PAdapter_Config;

#define HD_STATUS_FREE                          ARC_BIT(0)
#define HD_STATUS_ASSIGNED                      ARC_BIT(1)
#define HD_STATUS_SPARE                         ARC_BIT(2)
#define HD_STATUS_OFFLINE                       ARC_BIT(3)
#define HD_STATUS_SMARTCHECKING                 ARC_BIT(4)
#define HD_STATUS_MP                            ARC_BIT(5)

#define HD_BGA_TYPE_NONE						0
#define HD_BGA_TYPE_MP							1
#define HD_BGA_TYPE_DATASCRUB					2

#define HD_STATUS_SETONLINE                     0
#define HD_STATUS_SETOFFLINE                    1
#define HD_STATUS_INVALID                       0xFF

//HDType
#define HD_TYPE_SATA                            ARC_BIT(0)
#define HD_TYPE_PATA                            ARC_BIT(1)
#define HD_TYPE_SAS                             ARC_BIT(2)
#define HD_TYPE_ATAPI                           ARC_BIT(3)
#define HD_TYPE_TAPE                            ARC_BIT(4)
#define HD_TYPE_SES                             ARC_BIT(5)
#define HD_TYPE_MEDIA_CHANGER                   ARC_BIT(6)	

#define HD_SPEED_1_5G							1
#define HD_SPEED_3G								2

// PIOMode
#define HD_PIO_NONE                             0
#define HD_PIO_1                                1
#define HD_PIO_2                                2
#define HD_PIO_3                                3
#define HD_PIO_4                                4
#define HD_PIO_5                                5

#define HD_DMA_NONE                             0
#define HD_DMA_1                                1
#define HD_DMA_2                                2
#define HD_DMA_3                                3
#define HD_DMA_4                                4
#define HD_DMA_5                                5
#define HD_DMA_6                                6
#define HD_DMA_7                                7
#define HD_DMA_8                                8
#define HD_DMA_9                                9

// ConnectionType
#define DC_ATA                                      ARC_BIT(0)
#define DC_SCSI                                     ARC_BIT(1)
#define DC_SERIAL                                   ARC_BIT(2)
#define DC_PARALLEL                                 ARC_BIT(3)
#define DC_ATAPI                                    ARC_BIT(4)
#define DC_SGPIO                                    ARC_BIT(5)

// DeviceType
#define DT_DIRECT_ACCESS_BLOCK                      0x00
#define DT_SEQ_ACCESS                               0x01
#define DT_PRINTER                                  0x02
#define DT_PROCESSOR                                0x03
#define DT_WRITE_ONCE                               0x04
#define DT_CD_DVD                                   0x05
#define DT_OPTICAL_MEMORY                           0x07
#define DT_MEDIA_CHANGER                            0x08
#define DT_STORAGE_ARRAY_CTRL                       0x0C
/* an actual enclosure */
#define DT_ENCLOSURE                                0x0D
/* The following are defined by Marvell */
#define DT_EXPANDER                                 0x20
#define DT_PM                                       0x21
/* a device that provides some SES services (ie, Luigi) */
#define DT_SES_DEVICE                               0x22
#define DT_LOGICAL_UNIT_NOT_PRESENT_DEVICE          0x7F

//FeatureSupport
#define HD_FEATURE_NCQ                          ARC_BIT(0)	/* Capability */
#define HD_FEATURE_TCQ                          ARC_BIT(1)
#define HD_FEATURE_1_5G                         ARC_BIT(2)
#define HD_FEATURE_3G                           ARC_BIT(3)
#define HD_FEATURE_6G                           ARC_BIT(4)
#define HD_FEATURE_WRITE_CACHE                  ARC_BIT(5)
#define HD_FEATURE_48BITS                       ARC_BIT(6)
#define HD_FEATURE_SMART                        ARC_BIT(7)
#define HD_FEATURE_DATA_SET_MAMAGEMENT          ARC_BIT(8)


// SesElementType
#define SES_TYPE_UNSPECIFIED                    0x00
#define SES_TYPE_DISK_DEVICE                    0x01
#define SES_TYPE_POWER_SUPPLY                   0x02
#define SES_TYPE_FAN                            0x03
#define SES_TYPE_TEMPERATURE_SENSOR             0x04
#define SES_TYPE_DOOR_LOCK                      0x05
#define SES_TYPE_SPEAKER                        0x06
#define SES_TYPE_ES_CONTROLLER                  0x07/* Enclosure SCC */
#define SES_TYPE_SCC_CONTROLLER                 0x08/* SCC */
#define SES_TYPE_NONVOLATILE_CACHE              0x09
#define SES_TYPE_UPS                            0x0B
#define SES_TYPE_DISPLAY                        0x0C
#define SES_TYPE_KEYPAD                         0x0D
#define SES_TYPE_ENCLOSURE                      0x0E
#define SES_TYPE_SCSI_TRANSCEIVER               0x0F
#define SES_TYPE_LANGUAGE                       0x10
#define SES_TYPE_COMM_PORT                      0x11
#define SES_TYPE_VOLTAGE_SENSOR                 0x12
#define SES_TYPE_CURRENT_SENSOR                 0x13
#define SES_TYPE_SCSI_TARGET_PORT               0x14
#define SES_TYPE_SCSI_INITIATOR_PORT            0x15
#define SES_TYPE_SIMPLE_SUBENCLOSURE            0x16
#define SES_TYPE_ARRAY_DEVICE                   0x17
#define SES_TYPE_VENDOR_SPECIFIC                0x80

typedef struct _HD_Info
{
	Link_Entity  Link;						/* 0x00: Including self DevID & DevType */
	U8           AdapterID;					/* 0x40: */
	U8           Status;					/* 0x41: Refer to HD_STATUS_XXX */
	U8           HDType;					/* 0x42: HD_Type_xxx, replaced by new driver with ConnectionType & DeviceType */
	U8           PIOMode;					/* 0x43: Max PIO mode */
	U8           MDMAMode;					/* 0x44: Max MDMA mode */
	U8           UDMAMode;					/* 0x45: Max UDMA mode */
	U8           ConnectionType;			/* 0x46: DC_XXX, ConnectionType & DeviceType in new driver to replace HDType above */
	U8           DeviceType;				/* 0x47: DT_XXX */

	U32          FeatureSupport;			/* 0x48: Support 1.5G, 3G, TCQ, NCQ, and etc, ARC_BIT related */
	U8           Model[40];					/* 0x4C: */
	U8           SerialNo[20];				/* 0x74: */
	U8           FWVersion[8];				/* 0x88: */
	U64          Size;						/* 0x90: unit: 1KB */
	U8           WWN[8];					/* 0x98: ATA/ATAPI-8 has such definitions for the identify buffer */
	U8           CurrentPIOMode;			/* 0xA0: Current PIO mode */
	U8           CurrentMDMAMode;			/* 0xA1: Current MDMA mode */
	U8           CurrentUDMAMode;			/* 0xA2: Current UDMA mode */
	U8           SesOverallElementIndex;    /* 0xA3: */
	U32          BlockSize;					/* 0xA4: */
	U8           ActivityLEDStatus;			/* 0xA8: */
	U8           LocateLEDStatus;			/* 0xA9: */
	U8           ErrorLEDStatus;			/* 0xAA: */
	U8           SesElementType;			/* 0xAB: */
	U8           SesSlotNumber;				/* 0xAC: */
	U8           HD_SSD_Type;				/* 0xAD: */
	U16          SesEntryIndex;				/* 0xAE: */
	U16          Rotation_Rate;             /* 0xB0: */
	U8           Reserved1[70];				/* 0xB2: */
}HD_Info, *PHD_Info;						/* total 248 bytes */


typedef struct _Exp_Info
{
	Link_Entity    Link;					/* 0x00: Including self DevID & DevType */
	U8             AdapterID;               /* 0x40: */
	BOOLEAN        Configuring;      
	BOOLEAN        RouteTableConfigurable;
	U8             PhyCount;
	U16            ExpChangeCount;
	U16            MaxRouteIndexes;
	U8             VendorID[8+1];
	U8             ProductID[16+1];
	U8             ProductRev[4+1];
	U8             ComponentVendorID[8+1];
	U16            ComponentID;
	U8             ComponentRevisionID;
	U8             Reserved1[17];
}Exp_Info, * PExp_Info;						/* total 132 bytes */


// DriveSpeed
#define SD_SPEED_1_5G							1
#define SD_SPEED_3G								2
#define SD_SPEED_6G								3
typedef struct _HD_Config
{  
	BOOLEAN        WriteCacheOn;   /* 1: enable write cache */
	BOOLEAN		   SMARTOn;        /* 1: enable S.M.A.R.T */
	BOOLEAN		   Online;         /* 1: to set HD online */
	U8             DriveSpeed;	  /* For SATA & SAS.  SD_SPEED_1_5G, SD_SPEED_3G, SD_SPEED_6G etc*/
	U8             Reserved[2];
	U16            ScsiDeviceId;
}HD_Config, *PHD_Config;

typedef struct  _HD_SmartStatus
{
	BOOLEAN        SmartThresholdExceeded;        
	U8             CCB_Status;
	U16            ScsiDeviceId;
}HD_SmartStatus, *PHD_SmartStatus;

typedef  struct _PM_Info
{
	Link_Entity    Link;					/* Including self DevID & DevType */
	U8             AdapterID;
	U8             ProductRevision;
	U8             PMSpecRevision;			/* 10 means 1.0, 11 means 1.1 */
	U8             NumberOfPorts;
	U16            VendorId;
	U16            DeviceId;
	U8             Reserved1[8];
}PM_Info, *PPM_Info;						/* total 80 bytes */


typedef struct _EVENT_CODE
{
	DWORD code:16;
	DWORD facility:13;	
	DWORD custcode:1;
	DWORD severity:2;
}sEVENT_CODE, *pEVENT_CODE;


// Define the facility codes
//
#define STATUS_FACILITY_SYSTEM           0x0
#define STATUS_FACILITY_RUNTIME          0x2
#define STATUS_FACILITY_STUBS            0x3
#define STATUS_FACILITY_IO_ERROR_CODE    0x4


//
// Define the severity codes
//
#define STATUS_SEVERITY_SUCCESS          0x0
#define STATUS_SEVERITY_INFORMATIONAL    0x1
#define STATUS_SEVERITY_WARNING          0x2
#define STATUS_SEVERITY_ERROR            0x3


typedef struct _EVENT_ELEMENT   /*32BYTES*/
{
    U32                 TimeStamp;                  /*  4bytes: y.d.h.m.m.s, search sTIMESTAMP */
    U16                 DeviceID;                   /*  2bytes: scsi id, scsi lun */ 
    U8                  EventID;                    /*  1bytes: 0...31*/
    U8                  EventType;                  /*  1bytes: slot type */
#define ARCSAS_EVENT_TYPE_MESSAGE_ONLY              0
#define ARCSAS_EVENT_TYPE_PM_SLOT                   1
#define ARCSAS_EVENT_TYPE_EXPANDER_SLOT             2
#define ARCSAS_EVENT_TYPE_HBA_SLOT                  3
#define ARCSAS_EVENT_TYPE_PM                        4
#define ARCSAS_EVENT_TYPE_EXPANDER                  5
#define ARCSAS_EVENT_TYPE_ENCLOSURE                 6
#define ARCSAS_EVENT_TYPE_HBA                       7

	U32              ErrorCode;                  /*  4bytes: */
    U16              ParentIndexChildSlot;       /*  2bytes: Parent,Child */
    U16              EventFlag;                  /*  2bytes: */
#define ARCSAS_EVENT_FLAG_ERROR_INFO_RECORD         ARC_BIT(0)   /*EventScratchPad:"ERR_INFO_RECORD"             */
#define	ARCSAS_EVENT_FLAG_DEVICE_ARRIVAL            ARC_BIT(1)   /*EventScratchPad:Enc#1 Slot#1,Pm#1 Dev#1, Dev#1*/
#define	ARCSAS_EVENT_FLAG_DEVICE_REMOVAL            ARC_BIT(2)   /*EventScratchPad:Enc#1 Slot#1,Pm#1 Dev#1, Dev#1*/
#define ARCSAS_EVENT_FLAG_DEVICE_FAILED             ARC_BIT(3)   /*EventScratchPad:Enc#1 Slot#1,Pm#1 Dev#1, Dev#1*/
#define ARCSAS_EVENT_FLAG_DEVICE_READING_ERROR      ARC_BIT(4)   /*EventScratchPad:Enc#1 Slot#1,Pm#1 Dev#1, Dev#1*/
#define ARCSAS_EVENT_FLAG_HW_FAILED                 ARC_BIT(5)   /*EventScratchPad:RTC Battery,...               */
#define ARCSAS_EVENT_FLAG_RAID                      ARC_BIT(6)   /*EventScratchPad:CREATE VOLUME                 */
                                                                 /*EventScratchPad:DELETE VOLUME                 */
                                                                 /*EventScratchPad:ABORT REBUILDING              */
                                                                 /*EventScratchPad:START REBUILDING              */
                                                                 /*EventScratchPad:COMPLETE REBUILD              */
                                                                 /*EventScratchPad:ABORT INIT                    */
                                                                 /*EventScratchPad:START INIT                    */
                                                                 /*EventScratchPad:COMPLETE INIT                 */
                                                                 /*EventScratchPad:STOP INIT                     */
                                                                 /*EventScratchPad:VOLUME DEGRADED               */

    U8                EventScratchPad[16];        /* 16bytes: scratch pad for EventFlag description*/
}sEVENT_ELEMENT, *pEVENT_ELEMENT;


typedef struct _DRV_MAP_SAS
{
	int total;
	struct
	{
		HD_Info info;
		int luns;       /* # of LUN devices */
	}drv[ARCSAS_MAX_HDD_ELEMENT];
}sDRV_MAP_SAS, *pDRV_MAP_SAS;


typedef struct _EXP_MAP_SAS
{
	int total;
	struct
	{
		Exp_Info expInfo;
	}exp[ARCSAS_MAX_EXPANDER_SUPPORTED];
}sEXP_MAP_SAS, *pEXP_MAP_SAS;


typedef struct _SAS_ATA_PASSTHROUGH_COMMAND
{
	struct
	{   
		U8 features_error;   /* [in]: features [out]: error */
		U8 sector_count;	 /* [in] */
		U8 sector_number;	 /* [in] */
		U8 cylinder_low;	 /* [in/out] */
		U8 cylinder_high;    /* [in/out] */
		U8 device_head;      /* [in] */
		U8 command_status;   /* [in]: command [out]: status */
	}regs;
	U8 command_mode;		 /* [in], PIO Data-In / PIO Data-Out / Non-Data */	
	U8 data[512];			 /* [in/out] buffer for outgoing/incoming data */
	U32 datalen;
} sATA_PT_CMD, *pATA_PT_CMD;


typedef struct _SAS_SCSI_PASSTHROUGH_COMMAND
{	
	U8 cdb[CDB_LEN];       /* [in] 16 bytes CDB */		
	U8 command_mode;	   /* [in], PIO Data-In / PIO Data-Out / Non-Data */
	void *data;			   /* [in/out] buffer for outgoing/incoming data */
	U32 datalen;
	U8 reserved[7];
}sSCSI_PT_CMD, *pSCSI_PT_CMD;


#ifdef _WIN32
class ARCLIB_API CArclibSAS : public CArclibBase
{
#else
class CArclibSAS : public CArclibBase
{		
#endif


public:
	CArclibSAS();
	~CArclibSAS();

public:
	/************************************************************************************
	Function Name:	ArcGetLibVersion
	Function:		Get the areca library version
	Input:			None.
	Output:			None.
	Return Values:	1. Library version
	Note:			None.
	************************************************************************************/
	static int SASGetLibVersion
	(
		void
	);
	
	virtual ArcInterface *ArcGetInterface();
	/************************************************************************************
	Function Name:	ArcInitSession
	Function:		Startup areca library by assigning an interface to it
	Input:			1. ArcInterface
	Output:			None.
	Return Values:	1. ARC_STATUS					
	Note:			This function should be called before any member functions be called
	************************************************************************************/	
	virtual ARC_STATUS ArcInitSession
	(
		IN ArcInterface *Interface
	);

public:		
	/************************************************************************************
	Function Name:	ArcSetVendorString
	Function:		Set vendor string of HBA
	Input:			1. Vendor string(null-terminated), Max 9 characters
	Output:			None.
	Return Values:	1. ARC_STATUS
	Note:			None.
	************************************************************************************/
	virtual ARC_STATUS ArcSetVendorString
	(
		IN char *szVendorName
	);

	/************************************************************************************
	Function Name:	ArcSetModelString
	Function:		Set model string of HBA
	Input:			1. Model string(null-terminated), Max 10 characters
	Output:			None.
	Return Values:	1. ARC_STATUS
	Note:			None.
	************************************************************************************/
	virtual ARC_STATUS ArcSetModelString
	(
		IN char *szModelName
	);


	/************************************************************************************
	Function Name:	ArcGetEnclosureProperty
	Function:		Get the Enclosure property
	Input:			1. nId: Enclosure ID ( From ARCSAS_MIN_ENCLOSURE_ID to ARCSAS_MAX_ENCLOSURE_ID )
	                2. nLun: Enclosure LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )
	Output:			1. Pointer to a sGUI_ENC_PROPERTY structure that receives the information.
	Return Values:	ARC_STATUS
	Note:
	************************************************************************************/
	virtual ARC_STATUS ArcGetEnclosureProperty
	(		
		OUT pGUI_ENC_PROPERTY pEncProperty,
		IN BYTE nId, 
		IN BYTE nLun = 0 
	);

	/************************************************************************************
	Function Name:	ArcGetEnclosureDescriptor
	Function:		Get the Enclosure descriptor
	Input:			1. nItem: see 'enum _SAS_HW' 
					2. nBufsize: the size of output buffer
					3. nId: Enclosure ID ( From ARCSAS_MIN_ENCLOSURE_ID to ARCSAS_MAX_ENCLOSURE_ID )
	                4. nLun: Enclosure LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )
	Output:			1. pDescbuf: Pointer to a buffer that receives the item descriptor.
	Return Values:	ARC_STATUS
	Note:
	************************************************************************************/	
	virtual ARC_STATUS ArcGetEnclosureDescriptor
	(		
		IN BYTE nItem, 		
		IN int nBufsize,
		OUT BYTE *pDescbuf, 
		IN BYTE nId,
		IN BYTE nLun = 0
	);


	/************************************************************************************
	Function Name:	ArcGetEnclosureHwMonitor
	Function:		Get the Enclosure H/W monitor
	Input:			1. nId: Enclosure ID ( From ARCSAS_MIN_ENCLOSURE_ID to ARCSAS_MAX_ENCLOSURE_ID )
	                2. nLun: Enclosure LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )
	Output:			1. Pointer to a sSENSORS structure that receives the information.
	Return Values:	ARC_STATUS
	Note:
	************************************************************************************/	
	virtual ARC_STATUS ArcGetEnclosureHwMonitor
	(		
		OUT pSENSORS pHwInfo,
		IN BYTE nId, 
		IN BYTE nLun = 0
	);

	/************************************************************************************
    Function Name:	ArcFlashDisk
    Function:		Flash the disk
    Input:          None.
    Output:			None.  	
    Return Values:	1. ARC_STATUS
    Note:				                
	************************************************************************************/
	virtual ARC_STATUS ArcRemapDisk
	(
	);

	/************************************************************************************
	Function Name:	ArcGetPhysicalDriveInfo ( same as SASGetHDInfo(...) )
	Function:		Get the physical drive information
	Input:			1. nId: Hdd's ID ( From ARCSAS_MIN_HDD_ID to ARCSAS_MAX_HDD_ID )
	                2. nLun: Hdd's LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )
	Output:			1. Pointer to a HD_Info structure that receives the information.
	Return Values:	1. ARC_STATUS
	Note:			
	************************************************************************************/
	virtual ARC_STATUS ArcGetPhysicalDriveInfo
	(	
		OUT	HD_Info *info,
		IN BYTE nId,		
		IN BYTE nLun = 0	
	);

	/************************************************************************************
    Function Name:	ArcIdentifyDisk
    Function:		Identify the disk
    Input:			1. nMode: 0->Indicator on, 2->Indicator off
	                2. nId: Hdd's ID ( From ARCSAS_MIN_HDD_ID to ARCSAS_MAX_HDD_ID )
	                3. nLun: Hdd's LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )
    Output:			None.  	
    Return Values:	1. ARC_STATUS
    Note:				                
	************************************************************************************/
	virtual ARC_STATUS ArcIdentifyDisk
	(
		IN BYTE nMode, 
		IN BYTE nId, 
		IN BYTE nLun = 0
	);

	/************************************************************************************
    Function Name:	ArcIdentifyEnclosure
    Function:		Identify the enclosure
    Input:			1. nMode: 0->Indicator on, 2->Indicator off
	                2. nId: Enclosure ID ( From ARCSAS_MIN_ENCLOSURE_ID to ARCSAS_MAX_ENCLOSURE_ID )
	                3. nLun: Enclosure LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )
    Output:			None.  	
    Return Values:	1. ARC_STATUS
    Note:				                
	************************************************************************************/
	virtual ARC_STATUS ArcIdentifyEnclosure
	(		
		IN BYTE nMode, 
		IN BYTE nId,
		IN BYTE nLun = 0
	);	

	/************************************************************************************
    Function Name:	ArcSetDiskCacheMode
    Function:	    Set disk cache mode
    Input:          1. nMode: Mode of disk cache
	                2. nId: Hdd's ID ( From ARCSAS_MIN_HDD_ID to ARCSAS_MAX_HDD_ID )
	                3. nLun: Hdd's LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )	                
    Output:			None.  	
    Return Values:	1. ARC_STATUS
    Note:				                
	************************************************************************************/
	virtual ARC_STATUS ArcSetDiskCacheMode
	(
		IN BYTE nMode,	
		IN BYTE nId,
		IN BYTE nLun = 0
	);

	/************************************************************************************
    Function Name:	ArcSetDiskSmartMode
    Function:	    Set disk smart mode
    Input:          1. nMode: Mode of disk S.M.A.R.T
	                2. nId: Hdd's ID ( From ARCSAS_MIN_HDD_ID to ARCSAS_MAX_HDD_ID )
	                3. nLun: Hdd's LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )	
    Output:			None.  	
    Return Values:	1. ARC_STATUS
    Note:				                
	************************************************************************************/
	virtual ARC_STATUS ArcSetDiskSmartMode
	(
		IN BYTE nMode, 
		IN BYTE nId,
		IN BYTE nLun = 0
	);

	/************************************************************************************
    Function Name:	ArcSetDiskSpeedMode
    Function:	    Set disk speed mode
    Input:          1. nMode: Mode of disk speed
	                2. nId: Hdd's ID ( From ARCSAS_MIN_HDD_ID to ARCSAS_MAX_HDD_ID )
	                3. nLun: Hdd's LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )	
    Output:			None.  	
    Return Values:	1. ARC_STATUS
    Note:				                
	************************************************************************************/
	virtual ARC_STATUS ArcSetDiskSpeedMode
	(
		IN BYTE nMode,
		IN BYTE nId,
		IN BYTE nLun = 0
	);

	/************************************************************************************
    Function Name:	ArcSetDiskId
    Function:	    Set disk scsi id
    Input:          1. nNewId: New SCSI ID
	                2. nNewLun: New LUN
	                3. nId: Hdd's ID ( From ARCSAS_MIN_HDD_ID to ARCSAS_MAX_HDD_ID )
	                4. nLun: Hdd's LUN ( From ARCSAS_MIN_LUN To ARCSAS_MAX_LUN )	
    Output:			None.  	
    Return Values:	1. ARC_STATUS
    Note:				                
	************************************************************************************/
	virtual ARC_STATUS ArcSetDiskId
	(
		IN BYTE nNewId,
		IN BYTE nNewLun,
		IN BYTE nId,
		IN BYTE nLun = 0
	);		

	/************************************************************************************
	Function Name:	ArcClearEvent
	Function:		Clear all event buffer
	Input:			None.
	Output:			None.
	Return Values:	1. ARC_STATUS
	Note:			1. 
	************************************************************************************/
	virtual ARC_STATUS ArcClearEvent
	( 
		void 
	);

	/************************************************************************************
	Function Name:	ArcGetEventObject
	Function:		Get the individual event object
	Input:			1. Event object index 
	Output:			1. Pointer to an sEVENT_ELEMENT to save the current event object
	Return Values:	1. boolean value, > 0 means successed
	Note:			1. 
	************************************************************************************/
	virtual BOOL ArcGetEventObject
	(
		IN	int nEvtNum, 
		OUT	pEVENT_ELEMENT pEventObject
	);

	/************************************************************************************
	Function Name:	ArcPollEvent
	Function:		Get the current index of event object in the RAID subsystem
	Input:			None.
	Output:			1. Pointer to an INTEGER to save the index of current event object
	Return Values:	1. ARC_STATUS
	Note:
	************************************************************************************/
	virtual ARC_STATUS ArcPollEvent
	(
		OUT int *nCurrentEvtIndex
	);

	/************************************************************************************
	Function Name:	ArcGetReqEventPage
	Function:		Get the specific number of event pages from RAID subsystem
	Input:			1. The flag indicates how to retrieve the event object(s), 
	                   this parameter can be one of the following values:
					   1a. FLAG_ALL, retrieve all event objects inside the RAID subsystem
					   1b. FLAG_NEW, retrieve only new event objects
	Output:			1. Number of event objects retrieved
	Return Values:	1. ARC_STATUS	
	Note:			None.
	************************************************************************************/
	virtual ARC_STATUS ArcGetReqEventPage
	(
		IN	char bFlag, 	
		OUT int *iCount
	);	
	
	/************************************************************************************
	Function Name:	ArcUpdateFirmware
	Function:		Update firmware
	Input:			1. The path that point to a firmware file				
	Output:			None
	Return Values:	1. ARC_STATUS
	Note:
	************************************************************************************/
	virtual ARC_STATUS ArcUpdateFirmware
	(
		IN char *szFilePath
	);	


	
	enum
	{
		R_SMART_DATA = 0,		// (O)Read smart data
		R_THLD_DATA,			// (O)Read threshold data
		ED_AUTOSAVE,			// (X)Enable/disable autosave
		S_ATTR,					// (X)Save attribute
		EXE_OFFLINE,			// (O)Execute offline
		R_LOG,					// (O)Read log
		W_LOG,					// (X)Write log		
		E_SMART,				// (O)Enable SMART
		D_SMART,				// (O)Disable SMART
		RET_SMART_STATUS,		// (O)Return SMART status
		ED_AUTO_OFFLINE,		// (X)Enable/Disable automatic offline
		IDENTIFY,               // (O)Identify device
		CHECK_POWERMODE         // (O)Check power mode
	};

	/************************************************************************************
	 Function: ArcExeSMART
	 Parameters:  (i) cmd: the commands listed above
	              (i) nEncnum: enclosure#
				  (i) nDrvnum: drive#
				  (i) params: The options for EXE_OFFLINE, R_LOG only.
	                          For EXE_OFFLINE: SHORT_SELF_TEST indicates short self-test
					                           EXTENDED_SELF_TEST indicates extended self-test								
							  For R_LOG: 
	              (o) data: The option for R_SMART_DATA, R_THLD_DATA only.
							  the SMART data returned from the drive that indicates by drv, 
						      the memory allocation of 'data' is the responsibility of the
							  calling thread, and it's size should be >= 512 bytes		
	************************************************************************************/
	virtual ARC_STATUS ArcExeSMART
	(
	    IN BYTE nCmd,		
		IN BYTE nParams, 
		PBYTE pData, /* 512 bytes */
		IN BYTE nId,
		IN BYTE nLun = 0
	);

// ATA commands
#define ATA_IDENTIFY_DEVICE             0xec                                              
#define ATA_IDENTIFY_PACKET_DEVICE      0xa1
#define ATA_SMART_CMD                   0xb0
#define ATA_CHECK_POWER_MODE            0xe5

// ATA feature register values (ATA_SMART_CMD subcommands).
#define ATA_SMART_READ_VALUES           0xd0
#define ATA_SMART_READ_THRESHOLDS       0xd1
#define ATA_SMART_AUTOSAVE              0xd2
#define ATA_SMART_SAVE                  0xd3
#define ATA_SMART_IMMEDIATE_OFFLINE     0xd4
#define ATA_SMART_READ_LOG_SECTOR       0xd5
#define ATA_SMART_WRITE_LOG_SECTOR      0xd6
//#define ATA_SMART_WRITE_THRESHOLDS      0xd7
#define ATA_SMART_ENABLE                0xd8
#define ATA_SMART_DISABLE               0xd9
#define ATA_SMART_STATUS                0xda
#define ATA_SMART_AUTO_OFFLINE          0xdb

// Sector number values for ATA_SMART_IMMEDIATE_OFFLINE subcommand
#define OFFLINE_FULL_SCAN               0
#define SHORT_SELF_TEST                 1
#define EXTEND_SELF_TEST                2
#define CONVEYANCE_SELF_TEST            3
#define SELECTIVE_SELF_TEST             4
#define ABORT_SELF_TEST                 127
#define SHORT_CAPTIVE_SELF_TEST         129
#define EXTEND_CAPTIVE_SELF_TEST        130
#define CONVEYANCE_CAPTIVE_SELF_TEST    131
#define SELECTIVE_CAPTIVE_SELF_TEST     132
#define CAPTIVE_MASK                    (0x01<<7)

	virtual ARC_STATUS ArcAtaPassthroughCmd
	(		
		IN sATA_PT_CMD *pAtacmd,
		IN BYTE nId,
		IN BYTE nLun = 0
	);

	virtual ARC_STATUS ArcSCSIPassthroughCmd
	(		
		IN sSCSI_PT_CMD *pScsicmd,
		IN BYTE nId,
		IN BYTE nLun = 0
	);

	virtual ARC_STATUS SASGetHBAInfo(Adapter_Info *info);
	virtual ARC_STATUS SASGetHBAConfig(Adapter_Config *config);
	virtual ARC_STATUS SASSetHBAConfig(Adapter_Config *config);
	/************************************************************************************
	Function Name:	ArcUpdateFlashData
	Function:		Update firmware
	Input:			1. The buffer contains the data to be updated
					2. The size of data
	Output:			None
	Return Values:	1. ARC_STATUS
	Note:
	************************************************************************************/
	virtual ARC_STATUS ArcUpdateFlashData
	(
		IN char *pData, 
		IN int nDataSize,
		IN int nDataOffset
	);	

	virtual ARC_STATUS ArcReDiscovery
	(
		void
	);	
		
	virtual ARC_STATUS SASGetHDInfo
	(
		OUT HD_Info *info, 
		IN BYTE nId,
		IN BYTE nLun = 0 
	);

	virtual ARC_STATUS SASGetHDLunInfo
	(
		OUT HD_Info *info, 
		IN BYTE nId,
		IN BYTE nLun = 0 
	);

	virtual ARC_STATUS ReportLuns(int scsiid, sDRV_MAP_SAS *drvMap);
	
	virtual ARC_STATUS SASGetHDConfig
	(
		OUT HD_Config *config, 
		IN BYTE nId,
		IN BYTE nLun = 0
	);
	
	virtual ARC_STATUS SASSetHDConfig
	(
		IN BYTE cgcode, 
		IN BYTE nId,
		IN BYTE nLun = 0
	);
	
	virtual ARC_STATUS SASGetHDSmartStatus
	(
		OUT HD_SmartStatus *ss, 
		IN BYTE nId,
		IN BYTE nLun = 0
	);
	
	virtual ARC_STATUS SASGetExpInfo
	(
		OUT Exp_Info *info, 
		IN BYTE nId,
		IN BYTE nLun = 0
	);
	
	virtual ARC_STATUS SASGetPMInfo
	(
		OUT PM_Info *info, 
		IN BYTE nId,
		IN BYTE nLun = 0
	);
	//
	// helper functions
	//	
	ARC_STATUS SASHelpBuildDeviceMap(sDRV_MAP_SAS *drvMap);
	ARC_STATUS SASHelpBuildExpanderMap(sEXP_MAP_SAS *raidMap);
	
protected:		
	BOOL	IsValidHandle();	
	ARC_STATUS SASCmdReadWrite(BYTE cmd, void *w_buf, int w_buf_len, void *r_buf, int r_buf_len, int &actual_read_size, DWORD flag = 0);
	
	ARC_STATUS ArcFlashErase(void);
	ARC_STATUS ArcFlashRead(IN BYTE *buf, IN ULONG offset);
	ARC_STATUS ArcFlashWrite(IN BYTE *buf, IN ULONG offset);
	ARC_STATUS ArcGetFlashSize(OUT ULONG *size);	

	typedef struct _TYPE_DESC_HDR_LIST
	{
		unsigned char type;
		unsigned char num_elements;
		unsigned char sub_enc_id;
		unsigned char text_len;
	}sTYPE_DESC_HDR_LIST, *pTYPE_DESC_HDR_LIST;

	ARC_STATUS ses_parse_configuration_page(int fd, void *resp, int mx_resp_len, int bus, int id, int lun, unsigned char *gen_code, sTYPE_DESC_HDR_LIST *type_hdr);
	ARC_STATUS scsi_inquiry(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	ARC_STATUS scsi_report_luns(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);	
	ARC_STATUS scsi_get_diagnostic_page(int page, void * resp, int mx_resp_len, int bus, int id, int lun);
	ARC_STATUS scsi_set_diagnostic_page(int page, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page00
	ARC_STATUS ses_get_supported_diagnostic_pages_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page01
	ARC_STATUS ses_get_configuration_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page02
	ARC_STATUS ses_get_enclosure_status_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	ARC_STATUS ses_set_enclosure_status_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page03
	ARC_STATUS ses_get_help_text_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page04
	ARC_STATUS ses_get_string_in_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page05
	ARC_STATUS ses_get_threshold_in_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page06
	ARC_STATUS ses_get_obsolete_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page07
	ARC_STATUS ses_get_element_descriptor_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page08
	ARC_STATUS ses_get_short_enclosure_status_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page09
	ARC_STATUS ses_get_enclosure_busy_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page0a
	ARC_STATUS ses_get_additional_enclosure_status_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page0b
	ARC_STATUS ses_get_subenclosure_help_text_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page0c
	ARC_STATUS ses_get_subenclosure_string_in_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page0d
	ARC_STATUS ses_get_supported_ses_diagnostic_pages_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page0e
	ARC_STATUS ses_get_download_microcode_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);
	// page0e
	ARC_STATUS ses_set_download_microcode_dp(int fd, void * resp, int mx_resp_len, int bus, int id, int lun);

private:	
	// save the previous index of event
	int		m_iPreIndex;
	// save the current index of event
	int		m_iCurrentIndex;
	int		m_iNewCount;
	BYTE	*m_pEventMapping;
	int		m_iMtu;		

};

#pragma pack()
#ifdef __cplusplus
}
#endif



#endif



